"""Tests of parsing SSCCs."""

import pytest

from biip import ParseError
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.sscc import Sscc


def test_parse() -> None:
    sscc = Sscc.parse("157035381410375177")

    assert sscc == Sscc(
        value="157035381410375177",
        prefix=GS1Prefix(value="570", usage="GS1 Denmark"),
        company_prefix=GS1CompanyPrefix(value="5703538"),
        extension_digit=1,
        payload="15703538141037517",
        check_digit=7,
    )


def test_parse_strips_surrounding_whitespace() -> None:
    sscc = Sscc.parse("  \t 157035381410375177 \n  ")

    assert sscc.value == "157035381410375177"


def test_parse_value_with_invalid_length() -> None:
    with pytest.raises(ParseError) as exc_info:
        Sscc.parse("123")

    assert (
        str(exc_info.value)
        == "Failed to parse '123' as SSCC: Expected 18 digits, got 3."
    )


@pytest.mark.parametrize("value", ["012345678901234abc", "012345678901234⁰⁰⁰"])
def test_parse_nonnumeric_value(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        Sscc.parse(value)

    assert (
        str(exc_info.value)
        == f"Failed to parse {value!r} as SSCC: Expected a numerical value."
    )


def test_parse_with_invalid_check_digit() -> None:
    with pytest.raises(ParseError) as exc_info:
        Sscc.parse("376130321109103421")

    assert (
        str(exc_info.value)
        == "Invalid SSCC check digit for '376130321109103421': Expected 0, got 1."
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("157035381410375177", "1 5703538 141037517 7"),
        ("357081300469846950", "3 5708130 046984695 0"),
        ("370595680445154697", "3 705956 8044515469 7"),
        # GS1 Prefix 671 is currently unassigned:
        ("376130321109103420", "3 7613032110910342 0"),
    ],
)
def test_as_hri(value: str, expected: str) -> None:
    sscc = Sscc.parse(value)

    assert sscc.as_hri() == expected


def test_as_hri_with_too_low_company_prefix_length() -> None:
    sscc = Sscc.parse("376130321109103420")

    with pytest.raises(
        ValueError,
        match=r"^Expected company prefix length between 7 and 10, got 6.$",
    ):
        sscc.as_hri(company_prefix_length=6)


def test_as_hri_with_too_high_company_prefix_length() -> None:
    sscc = Sscc.parse("376130321109103420")

    with pytest.raises(
        ValueError,
        match=r"^Expected company prefix length between 7 and 10, got 11.$",
    ):
        sscc.as_hri(company_prefix_length=11)



import pytest

from biip import ParseConfig
from biip.gtin import Gtin, GtinFormat
from biip.rcn import Rcn, RcnRegion, RcnUsage


@pytest.mark.parametrize(
    ("value", "gtin_format", "rcn_usage"),
    [
        # RCN-8
        ("00011112", GtinFormat.GTIN_8, RcnUsage.COMPANY),
        ("00099998", GtinFormat.GTIN_8, RcnUsage.COMPANY),
        # RCN-12
        ("201111111115", GtinFormat.GTIN_12, RcnUsage.GEOGRAPHICAL),
        ("291111111116", GtinFormat.GTIN_12, RcnUsage.GEOGRAPHICAL),
        ("401111111119", GtinFormat.GTIN_12, RcnUsage.COMPANY),
        ("491111111110", GtinFormat.GTIN_12, RcnUsage.COMPANY),
        # RCN-13
        ("2001111111119", GtinFormat.GTIN_13, RcnUsage.GEOGRAPHICAL),
        ("2991111111113", GtinFormat.GTIN_13, RcnUsage.GEOGRAPHICAL),
    ],
)
def test_gtin_parse_may_return_rcn_instance(
    value: str, gtin_format: GtinFormat, rcn_usage: RcnUsage
) -> None:
    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.SWEDEN))

    assert isinstance(rcn, Rcn)
    assert rcn.format == gtin_format
    assert rcn.usage == rcn_usage
    if rcn_usage == RcnUsage.GEOGRAPHICAL:
        assert rcn.region == RcnRegion.SWEDEN
    else:
        assert rcn.region is None


def test_rcn_without_specified_region() -> None:
    rcn = Gtin.parse("2991111111113", config=ParseConfig(rcn_region=None))

    assert isinstance(rcn, Rcn)
    assert rcn.format == GtinFormat.GTIN_13
    assert rcn.usage == RcnUsage.GEOGRAPHICAL
    assert rcn.region is None
    assert rcn.weight is None
    assert rcn.price is None
    assert rcn.money is None


def test_gtin_14_with_rcn_prefix_is_not_an_rcn() -> None:
    # The value below is a GTIN-14 composed of packaging level 1 and a valid RCN-13.
    gtin = Gtin.parse("12991111111110", config=ParseConfig(rcn_region=None))

    assert isinstance(gtin, Gtin)
    assert not isinstance(gtin, Rcn)
    assert gtin.format == GtinFormat.GTIN_14


@pytest.mark.parametrize(
    ("value", "rcn_region"),
    [
        ("de", RcnRegion.GERMANY),
        ("dk", RcnRegion.DENMARK),
        ("ee", RcnRegion.ESTONIA),
        ("fi", RcnRegion.FINLAND),
        ("gb", RcnRegion.GREAT_BRITAIN),
        ("lt", RcnRegion.LITHUANIA),
        ("lv", RcnRegion.LATVIA),
        ("no", RcnRegion.NORWAY),
        ("se", RcnRegion.SWEDEN),
    ],
)
def test_rcn_region_can_be_specified_as_string(
    value: str, rcn_region: RcnRegion
) -> None:
    rcn = Gtin.parse("0211111111114", config=ParseConfig(rcn_region=value))

    assert isinstance(rcn, Rcn)
    assert rcn.region == rcn_region


def test_fails_when_rcn_region_is_unknown_string() -> None:
    with pytest.raises(
        ValueError,
        match=r"^'foo' is not a valid RcnRegion$",
    ):
        Gtin.parse("2311111112345", config=ParseConfig(rcn_region="foo"))


def test_rcn_usage_repr() -> None:
    assert repr(RcnUsage.COMPANY) == "RcnUsage.COMPANY"


def test_rcn_region_repr() -> None:
    assert repr(RcnRegion.ESTONIA) == "RcnRegion.ESTONIA"



import pytest

from biip import ParseError
from biip.symbology import GS1Symbology, ISOSymbology, SymbologyIdentifier


def test_gs1_symbology_enum() -> None:
    assert GS1Symbology("C1") == GS1Symbology.GS1_128


def test_gs1_symbology_with_gs1_messages() -> None:
    assert GS1Symbology.EAN_13 not in GS1Symbology.with_gs1_messages()
    assert GS1Symbology.GS1_128 in GS1Symbology.with_gs1_messages()


def test_gs1_symbology_with_gs1_web_uri() -> None:
    assert GS1Symbology.GS1_DATAMATRIX in GS1Symbology.with_gs1_web_uri()
    assert GS1Symbology.GS1_QR_CODE in GS1Symbology.with_gs1_web_uri()


def test_gs1_symbology_with_gtin() -> None:
    assert GS1Symbology.EAN_13 in GS1Symbology.with_gtin()

    # Even though GS1-128 can contain a GTIN,
    # it cannot be parsed with a pure GTIN parser.
    assert GS1Symbology.GS1_128 not in GS1Symbology.with_gtin()


def test_iso_symbology_enum() -> None:
    assert ISOSymbology("C") == ISOSymbology.CODE_128


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "]E0abc",
            SymbologyIdentifier(
                value="]E0",
                iso_symbology=ISOSymbology.EAN_UPC,
                modifiers="0",
                gs1_symbology=GS1Symbology.EAN_13,
            ),
        ),
        (
            "]C0abc",
            SymbologyIdentifier(
                value="]C0",
                iso_symbology=ISOSymbology.CODE_128,
                modifiers="0",
                gs1_symbology=None,
            ),
        ),
        (
            "]C1abc",
            SymbologyIdentifier(
                value="]C1",
                iso_symbology=ISOSymbology.CODE_128,
                modifiers="1",
                gs1_symbology=GS1Symbology.GS1_128,
            ),
        ),
        (
            "]I1abc",
            SymbologyIdentifier(
                value="]I1",
                iso_symbology=ISOSymbology.ITF,
                modifiers="1",
                gs1_symbology=GS1Symbology.ITF_14,
            ),
        ),
        (
            "]Y1abcdefghijkl",
            SymbologyIdentifier(
                value="]Y1a",
                iso_symbology=ISOSymbology.SYSTEM_EXPANSION,
                modifiers="1a",
                gs1_symbology=None,
            ),
        ),
        (
            "]Y9abcdefghijkl",
            SymbologyIdentifier(
                value="]Y9abcdefghi",
                iso_symbology=ISOSymbology.SYSTEM_EXPANSION,
                modifiers="9abcdefghi",
                gs1_symbology=None,
            ),
        ),
    ],
)
def test_extract(value: str, expected: SymbologyIdentifier) -> None:
    assert SymbologyIdentifier.extract(value) == expected


def test_extract_fails_if_no_flag_character() -> None:
    with pytest.raises(ParseError) as exc_info:
        SymbologyIdentifier.extract("E0")

    assert str(exc_info.value) == (
        "Failed to get Symbology Identifier from 'E0'. "
        "No initial ']' flag character found."
    )


def test_extract_fails_if_unknown_code_character() -> None:
    with pytest.raises(ParseError) as exc_info:
        SymbologyIdentifier.extract("]J0")

    assert str(exc_info.value) == (
        "Failed to get Symbology Identifier from ']J0'. "
        "'J' is not a recognized code character."
    )


@pytest.mark.parametrize("value", ["]E0", "]C1", "]Y1a", "]Y9abcdefghi"])
def test_len(value: str) -> None:
    assert len(SymbologyIdentifier.extract(value)) == len(value)


@pytest.mark.parametrize("value", ["]E0", "]C1", "]Y1a", "]Y9abcdefghi"])
def test_str(value: str) -> None:
    assert str(SymbologyIdentifier.extract(value)) == value



import datetime as dt
from decimal import Decimal

import pytest

from biip import ParseConfig, ParseResult, parse
from biip.gs1_application_identifiers import GS1ApplicationIdentifier
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings
from biip.gs1_messages import GS1Message
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.gs1_web_uris import GS1WebURI
from biip.gtin import Gtin, GtinFormat
from biip.rcn import Rcn, RcnRegion, RcnUsage
from biip.sscc import Sscc
from biip.symbology import GS1Symbology, ISOSymbology, SymbologyIdentifier
from biip.upc import Upc, UpcFormat


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            # GTIN-8
            "96385074",
            ParseResult(
                value="96385074",
                gtin=Gtin(
                    value="96385074",
                    format=GtinFormat.GTIN_8,
                    prefix=GS1Prefix(value="00009", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0000963"),
                    payload="9638507",
                    check_digit=4,
                ),
                upc_error=(
                    "Invalid UPC-E number system for '96385074': "
                    "Expected 0 or 1, got 9."
                ),
                sscc_error=(
                    "Failed to parse '96385074' as SSCC: Expected 18 digits, got 8."
                ),
                gs1_message=GS1Message(
                    value="96385074",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("96"),
                                value="385074",
                                pattern_groups=["385074"],
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GTIN-8 with Symbology Identifier
            "]E496385074",
            ParseResult(
                value="]E496385074",
                symbology_identifier=SymbologyIdentifier(
                    value="]E4",
                    iso_symbology=ISOSymbology.EAN_UPC,
                    modifiers="4",
                    gs1_symbology=GS1Symbology.EAN_8,
                ),
                gtin=Gtin(
                    value="96385074",
                    format=GtinFormat.GTIN_8,
                    prefix=GS1Prefix(value="00009", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0000963"),
                    payload="9638507",
                    check_digit=4,
                ),
            ),
        ),
        (
            # Both a valid GTIN-8 and a valid UPC-E. GTIN result should be the
            # GTIN-8, not the UPC-E expanded to UPC-A/GTIN-12.
            "12345670",
            ParseResult(
                value="12345670",
                gtin=Gtin(
                    value="12345670",
                    format=GtinFormat.GTIN_8,
                    prefix=GS1Prefix(value="00001", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0000123"),
                    payload="1234567",
                    check_digit=0,
                ),
                upc=Upc(
                    value="12345670",
                    format=UpcFormat.UPC_E,
                    number_system_digit=1,
                    payload="1234567",
                    check_digit=0,
                ),
                sscc_error=(
                    "Failed to parse '12345670' as SSCC: Expected 18 digits, got 8."
                ),
                gs1_message_error=(
                    "Failed to match '12345670' with GS1 AI (12) pattern "
                    r"'^12(\d{2}(?:0\d|1[0-2])(?:[0-2]\d|3[01]))$'."
                ),
            ),
        ),
        (
            # GTIN-12. GTIN-12 is also valid as UPC-A.
            "123601057072",
            ParseResult(
                value="123601057072",
                gtin=Gtin(
                    value="123601057072",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                    company_prefix=None,
                    payload="12360105707",
                    check_digit=2,
                ),
                upc=Upc(
                    value="123601057072",
                    format=UpcFormat.UPC_A,
                    number_system_digit=1,
                    payload="12360105707",
                    check_digit=2,
                ),
                sscc_error=(
                    "Failed to parse '123601057072' as SSCC: "
                    "Expected 18 digits, got 12."
                ),
                gs1_message_error=(
                    "Failed to get GS1 Application Identifier from '7072'."
                ),
            ),
        ),
        (
            # GTIN-12 with Symbology Identifier. GTIN-12 is also valid as UPC-A.
            "]E00123601057072",
            ParseResult(
                value="]E00123601057072",
                symbology_identifier=SymbologyIdentifier(
                    value="]E0",
                    iso_symbology=ISOSymbology.EAN_UPC,
                    modifiers="0",
                    gs1_symbology=GS1Symbology.EAN_13,
                ),
                gtin=Gtin(
                    value="0123601057072",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                    company_prefix=None,
                    payload="12360105707",
                    check_digit=2,
                ),
                upc=Upc(
                    value="123601057072",
                    format=UpcFormat.UPC_A,
                    number_system_digit=1,
                    payload="12360105707",
                    check_digit=2,
                ),
            ),
        ),
        (
            # GTIN-13
            "5901234123457",
            ParseResult(
                value="5901234123457",
                gtin=Gtin(
                    value="5901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
                upc_error=(
                    "Failed to parse '5901234123457' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 13."
                ),
                sscc_error=(
                    "Failed to parse '5901234123457' as SSCC: "
                    "Expected 18 digits, got 13."
                ),
                gs1_message_error=(
                    "Failed to get GS1 Application Identifier from '5901234123457'."
                ),
            ),
        ),
        (
            # GTIN-13 with Symbology Identifier
            "]E05901234123457",
            ParseResult(
                value="]E05901234123457",
                symbology_identifier=SymbologyIdentifier(
                    value="]E0",
                    iso_symbology=ISOSymbology.EAN_UPC,
                    modifiers="0",
                    gs1_symbology=GS1Symbology.EAN_13,
                ),
                gtin=Gtin(
                    value="5901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
            ),
        ),
        (
            # GTIN-14
            "05901234123457",
            ParseResult(
                value="05901234123457",
                gtin=Gtin(
                    value="05901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
                upc_error=(
                    "Failed to parse '05901234123457' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 14."
                ),
                sscc_error=(
                    "Failed to parse '05901234123457' as SSCC: "
                    "Expected 18 digits, got 14."
                ),
                gs1_message_error=(
                    "Failed to get GS1 Application Identifier from '05901234123457'."
                ),
            ),
        ),
        (
            # GTIN-14 with Symbology Identifier
            "]I105901234123457",
            ParseResult(
                value="]I105901234123457",
                symbology_identifier=SymbologyIdentifier(
                    value="]I1",
                    iso_symbology=ISOSymbology.ITF,
                    modifiers="1",
                    gs1_symbology=GS1Symbology.ITF_14,
                ),
                gtin=Gtin(
                    value="05901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
            ),
        ),
        (
            # UPC-E. UPC-E is expanded to UPC-A and exposed as GTIN too.
            # Coincidentally it is also a valid GS1 Message.
            "425261",
            ParseResult(
                value="425261",
                gtin=Gtin(
                    value="042100005264",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="004", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0042100"),
                    payload="04210000526",
                    check_digit=4,
                    packaging_level=None,
                ),
                upc=Upc(
                    value="425261",
                    format=UpcFormat.UPC_E,
                    number_system_digit=0,
                    payload="0425261",
                    check_digit=4,
                ),
                sscc_error=(
                    "Failed to parse '425261' as SSCC: Expected 18 digits, got 6."
                ),
                gs1_message=GS1Message(
                    value="425261",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("425"),
                                value="261",
                                pattern_groups=["261"],
                                gtin=None,
                                sscc=None,
                                date=None,
                                decimal=None,
                                money=None,
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # SSCC
            "157035381410375177",
            ParseResult(
                value="157035381410375177",
                gtin_error=(
                    "Failed to parse '157035381410375177' as GTIN: "
                    "Expected 8, 12, 13, or 14 digits, got 18."
                ),
                upc_error=(
                    "Failed to parse '157035381410375177' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 18."
                ),
                sscc=Sscc(
                    value="157035381410375177",
                    prefix=GS1Prefix(value="570", usage="GS1 Denmark"),
                    company_prefix=GS1CompanyPrefix(value="5703538"),
                    extension_digit=1,
                    payload="15703538141037517",
                    check_digit=7,
                ),
                gs1_message_error=(
                    "Failed to match '157035381410375177' with GS1 AI (15) pattern "
                    r"'^15(\d{2}(?:0\d|1[0-2])(?:[0-2]\d|3[01]))$'."
                ),
            ),
        ),
        (
            # GS1 AI: SSCC
            "00157035381410375177",
            ParseResult(
                value="00157035381410375177",
                gtin_error=(
                    "Failed to parse '00157035381410375177' as GTIN: "
                    "Expected 8, 12, 13, or 14 digits, got 20."
                ),
                upc_error=(
                    "Failed to parse '00157035381410375177' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 20."
                ),
                sscc=Sscc(
                    value="157035381410375177",
                    prefix=GS1Prefix(value="570", usage="GS1 Denmark"),
                    company_prefix=GS1CompanyPrefix(value="5703538"),
                    extension_digit=1,
                    payload="15703538141037517",
                    check_digit=7,
                ),
                gs1_message=GS1Message(
                    value="00157035381410375177",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("00"),
                                value="157035381410375177",
                                pattern_groups=["157035381410375177"],
                                sscc=Sscc(
                                    value="157035381410375177",
                                    prefix=GS1Prefix(value="570", usage="GS1 Denmark"),
                                    company_prefix=GS1CompanyPrefix(value="5703538"),
                                    extension_digit=1,
                                    payload="15703538141037517",
                                    check_digit=7,
                                ),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI: Invalid SSCC
            "00376130321109103421",
            ParseResult(
                value="00376130321109103421",
                gtin_error=(
                    "Failed to parse '00376130321109103421' as GTIN: "
                    "Expected 8, 12, 13, or 14 digits, got 20."
                ),
                upc_error=(
                    "Failed to parse '00376130321109103421' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 20."
                ),
                sscc_error=(
                    "Invalid SSCC check digit for "
                    "'376130321109103421': Expected 0, got 1."
                ),
                gs1_message=GS1Message(
                    value="00376130321109103421",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("00"),
                                value="376130321109103421",
                                pattern_groups=["376130321109103421"],
                                sscc_error=(
                                    "Invalid SSCC check digit for "
                                    "'376130321109103421': Expected 0, got 1."
                                ),
                            ),
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI: GTIN-12. GTIN-12 is also valid as UPC-A.
            "0100123601057072",
            ParseResult(
                value="0100123601057072",
                gtin=Gtin(
                    value="00123601057072",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                    company_prefix=None,
                    payload="12360105707",
                    check_digit=2,
                    packaging_level=None,
                ),
                upc=Upc(
                    value="123601057072",
                    format=UpcFormat.UPC_A,
                    number_system_digit=1,
                    payload="12360105707",
                    check_digit=2,
                ),
                sscc_error=(
                    "Failed to parse '0100123601057072' as SSCC: "
                    "Expected 18 digits, got 16."
                ),
                gs1_message=GS1Message(
                    value="0100123601057072",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="00123601057072",
                                pattern_groups=["00123601057072"],
                                gtin=Gtin(
                                    value="00123601057072",
                                    format=GtinFormat.GTIN_12,
                                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                                    company_prefix=None,
                                    payload="12360105707",
                                    check_digit=2,
                                    packaging_level=None,
                                ),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI: GTIN-13
            "0105901234123457",
            ParseResult(
                value="0105901234123457",
                gtin=Gtin(
                    value="05901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
                upc_error=(
                    "Failed to parse '0105901234123457' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 16."
                ),
                sscc_error=(
                    "Failed to parse '0105901234123457' as SSCC: "
                    "Expected 18 digits, got 16."
                ),
                gs1_message=GS1Message(
                    value="0105901234123457",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="05901234123457",
                                pattern_groups=["05901234123457"],
                                gtin=Gtin(
                                    value="05901234123457",
                                    format=GtinFormat.GTIN_13,
                                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                                    company_prefix=None,
                                    payload="590123412345",
                                    check_digit=7,
                                ),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI: Invalid GTIN
            "0105901234123458",
            ParseResult(
                value="0105901234123458",
                gtin_error=(
                    "Invalid GTIN check digit for '05901234123458': Expected 7, got 8."
                ),
                upc_error=(
                    "Failed to parse '0105901234123458' as UPC: "
                    "Expected 6, 7, 8, or 12 digits, got 16."
                ),
                sscc_error=(
                    "Failed to parse '0105901234123458' as SSCC: "
                    "Expected 18 digits, got 16."
                ),
                gs1_message=GS1Message(
                    value="0105901234123458",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="05901234123458",
                                pattern_groups=["05901234123458"],
                                gtin_error=(
                                    "Invalid GTIN check digit for '05901234123458': "
                                    "Expected 7, got 8."
                                ),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI: best before date
            "15210527",
            ParseResult(
                value="15210527",
                gtin_error=(
                    "Invalid GTIN check digit for '15210527': Expected 4, got 7."
                ),
                upc_error=(
                    "Invalid UPC-E check digit for '15210527': Expected 6, got 7."
                ),
                sscc_error=(
                    "Failed to parse '15210527' as SSCC: Expected 18 digits, got 8."
                ),
                gs1_message=GS1Message(
                    value="15210527",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("15"),
                                value="210527",
                                pattern_groups=["210527"],
                                date=dt.date(2021, 5, 27),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI with GS1-128 Symbology Identifier
            "]C1010590123412345715210526",
            ParseResult(
                value="]C1010590123412345715210526",
                symbology_identifier=SymbologyIdentifier(
                    value="]C1",
                    iso_symbology=ISOSymbology.CODE_128,
                    modifiers="1",
                    gs1_symbology=GS1Symbology.GS1_128,
                ),
                gtin=Gtin(
                    value="05901234123457",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                    company_prefix=None,
                    payload="590123412345",
                    check_digit=7,
                ),
                gs1_message=GS1Message(
                    value="010590123412345715210526",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="05901234123457",
                                pattern_groups=["05901234123457"],
                                gtin=Gtin(
                                    value="05901234123457",
                                    format=GtinFormat.GTIN_13,
                                    prefix=GS1Prefix(value="590", usage="GS1 Poland"),
                                    company_prefix=None,
                                    payload="590123412345",
                                    check_digit=7,
                                ),
                            ),
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("15"),
                                value="210526",
                                pattern_groups=["210526"],
                                date=dt.date(2021, 5, 26),
                            ),
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 AI with GS1 DataBar Symbology Identifier
            "]e00100123456789012",
            ParseResult(
                value="]e00100123456789012",
                symbology_identifier=SymbologyIdentifier(
                    value="]e0",
                    iso_symbology=ISOSymbology.RSS_EAN_UCC_COMPOSITE,
                    modifiers="0",
                    gs1_symbology=GS1Symbology.GS1_DATABAR,
                ),
                gtin=Gtin(
                    value="00123456789012",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                    company_prefix=None,
                    payload="12345678901",
                    check_digit=2,
                ),
                upc=Upc(
                    value="123456789012",
                    format=UpcFormat.UPC_A,
                    number_system_digit=1,
                    payload="12345678901",
                    check_digit=2,
                ),
                gs1_message=GS1Message(
                    value="0100123456789012",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="00123456789012",
                                pattern_groups=["00123456789012"],
                                gtin=Gtin(
                                    value="00123456789012",
                                    format=GtinFormat.GTIN_12,
                                    prefix=GS1Prefix(value="012", usage="GS1 US"),
                                    company_prefix=None,
                                    payload="12345678901",
                                    check_digit=2,
                                ),
                            )
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 Web URI with GTIN-12, lot number, and expiration date
            "https://example.com/gtin/614141123452/lot/ABC123?15=250330",
            ParseResult(
                value="https://example.com/gtin/614141123452/lot/ABC123?15=250330",
                gtin=Gtin(
                    value="00614141123452",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="061", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0614141"),
                    payload="61414112345",
                    check_digit=2,
                ),
                upc=Upc(
                    value="614141123452",
                    format=UpcFormat.UPC_A,
                    number_system_digit=6,
                    payload="61414112345",
                    check_digit=2,
                ),
                gs1_web_uri=GS1WebURI(
                    value="https://example.com/gtin/614141123452/lot/ABC123?15=250330",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="00614141123452",
                                pattern_groups=["00614141123452"],
                                gtin=Gtin(
                                    value="00614141123452",
                                    format=GtinFormat.GTIN_12,
                                    prefix=GS1Prefix(value="061", usage="GS1 US"),
                                    company_prefix=GS1CompanyPrefix(value="0614141"),
                                    payload="61414112345",
                                    check_digit=2,
                                ),
                            ),
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("10"),
                                value="ABC123",
                                pattern_groups=["ABC123"],
                            ),
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("15"),
                                value="250330",
                                pattern_groups=[
                                    "250330",
                                ],
                                gln=None,
                                gln_error=None,
                                gtin=None,
                                gtin_error=None,
                                sscc=None,
                                sscc_error=None,
                                date=dt.date(2025, 3, 30),
                                datetime=None,
                                decimal=None,
                                money=None,
                            ),
                        ]
                    ),
                ),
            ),
        ),
        (
            # GS1 Web URI with GTIN-12, lot number, and expiration date with GS1
            # QR Code symbology identifier
            "]Q3https://example.com/gtin/614141123452/lot/ABC123?15=250330",
            ParseResult(
                value="]Q3https://example.com/gtin/614141123452/lot/ABC123?15=250330",
                symbology_identifier=SymbologyIdentifier(
                    value="]Q3",
                    iso_symbology=ISOSymbology.QR_CODE,
                    modifiers="3",
                    gs1_symbology=GS1Symbology.GS1_QR_CODE,
                ),
                gtin=Gtin(
                    value="00614141123452",
                    format=GtinFormat.GTIN_12,
                    prefix=GS1Prefix(value="061", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="0614141"),
                    payload="61414112345",
                    check_digit=2,
                ),
                upc=Upc(
                    value="614141123452",
                    format=UpcFormat.UPC_A,
                    number_system_digit=6,
                    payload="61414112345",
                    check_digit=2,
                ),
                gs1_message_error=(
                    "Failed to get GS1 Application Identifier from "
                    "'https://example.com/gtin/614141123452/lot/ABC123?15=250330'."
                ),
                gs1_web_uri=GS1WebURI(
                    value="https://example.com/gtin/614141123452/lot/ABC123?15=250330",
                    element_strings=GS1ElementStrings(
                        [
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("01"),
                                value="00614141123452",
                                pattern_groups=["00614141123452"],
                                gtin=Gtin(
                                    value="00614141123452",
                                    format=GtinFormat.GTIN_12,
                                    prefix=GS1Prefix(value="061", usage="GS1 US"),
                                    company_prefix=GS1CompanyPrefix(value="0614141"),
                                    payload="61414112345",
                                    check_digit=2,
                                ),
                            ),
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("10"),
                                value="ABC123",
                                pattern_groups=["ABC123"],
                            ),
                            GS1ElementString(
                                ai=GS1ApplicationIdentifier.extract("15"),
                                value="250330",
                                pattern_groups=[
                                    "250330",
                                ],
                                gln=None,
                                gln_error=None,
                                gtin=None,
                                gtin_error=None,
                                sscc=None,
                                sscc_error=None,
                                date=dt.date(2025, 3, 30),
                                datetime=None,
                                decimal=None,
                                money=None,
                            ),
                        ]
                    ),
                ),
            ),
        ),
    ],
)
def test_parse(value: str, expected: ParseResult) -> None:
    assert parse(value) == expected


def test_parse_rcn_with_ignored_invalid_check_digit() -> None:
    # NOTE: GTINs with variable weight usually only appear on consumer units,
    # and thus in barcodes like EAN-13, but we use a GS1-128 barcode (symbology
    # identifier `]C1`) in this example, so that we get to test the correct
    # passing of the `rcn_verify_variable_measure` parameter throughout all the
    # classes.

    result = parse(
        "]C10102824040005133",
        config=ParseConfig(
            rcn_region=RcnRegion.GERMANY,
            rcn_verify_variable_measure=False,
        ),
    )

    assert result == ParseResult(
        value="]C10102824040005133",
        symbology_identifier=SymbologyIdentifier(
            value="]C1",
            iso_symbology=ISOSymbology.CODE_128,
            modifiers="1",
            gs1_symbology=GS1Symbology.GS1_128,
        ),
        gtin=Rcn(
            value="02824040005133",
            format=GtinFormat.GTIN_13,
            prefix=GS1Prefix.extract("282"),
            company_prefix=None,
            payload="282404000513",
            check_digit=3,
            usage=RcnUsage.GEOGRAPHICAL,
            region=RcnRegion.GERMANY,
            weight=Decimal("0.513"),
        ),
        gs1_message=GS1Message(
            value="0102824040005133",
            element_strings=GS1ElementStrings(
                [
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("01"),
                        value="02824040005133",
                        pattern_groups=["02824040005133"],
                        gtin=Rcn(
                            value="02824040005133",
                            format=GtinFormat.GTIN_13,
                            prefix=GS1Prefix.extract("282"),
                            company_prefix=None,
                            payload="282404000513",
                            check_digit=3,
                            usage=RcnUsage.GEOGRAPHICAL,
                            region=RcnRegion.GERMANY,
                            weight=Decimal("0.513"),
                        ),
                    )
                ]
            ),
        ),
    )


def test_parse_strips_surrounding_whitespace() -> None:
    result = parse("  \t 5901234123457 \n  ")

    assert result.value == "5901234123457"
    assert result.gtin is not None
    assert result.gtin.value == "5901234123457"


def test_parse_strips_symbology_identifier() -> None:
    result = parse("]E05901234123457")

    assert result.value == "]E05901234123457"
    assert result.symbology_identifier is not None
    assert result.symbology_identifier.value == "]E0"
    assert result.gtin is not None
    assert result.gtin.value == "5901234123457"


def test_parse_with_separator_char() -> None:
    result = parse(
        "101313|15210526",
        config=ParseConfig(separator_chars=["|"]),
    )

    assert result.gs1_message is not None
    assert result.gs1_message.as_hri() == "(10)1313(15)210526"


def test_parse_invalid_data() -> None:
    result = parse("abc")

    assert result.gtin is None
    assert (
        result.gtin_error
        == "Failed to parse 'abc' as GTIN: Expected 8, 12, 13, or 14 digits, got 3."
    )
    assert result.gs1_message is None
    assert (
        result.gs1_message_error
        == "Failed to get GS1 Application Identifier from 'abc'."
    )
    assert result.gs1_web_uri is None
    assert result.gs1_web_uri_error is None
    assert result.sscc is None
    assert (
        result.sscc_error == "Failed to parse 'abc' as SSCC: Expected 18 digits, got 3."
    )
    assert result.upc is None
    assert (
        result.upc_error
        == "Failed to parse 'abc' as UPC: Expected 6, 7, 8, or 12 digits, got 3."
    )


def test_parse_invalid_http_uri() -> None:
    result = parse("https://example.com/foo/bar")

    assert result.gtin is None
    assert result.gtin_error is None
    assert result.gs1_message is None
    assert result.gs1_message_error is None
    assert result.gs1_web_uri is None
    assert (
        result.gs1_web_uri_error
        == "Expected a primary identifier in the path, got '/foo/bar'."
    )
    assert result.sscc is None
    assert result.sscc_error is None
    assert result.upc is None
    assert result.upc_error is None



"""Tests of parsing GTINs."""

import pytest

from biip import ParseError
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.gtin import Gtin, GtinFormat


def test_parse_value_with_invalid_length() -> None:
    with pytest.raises(ParseError) as exc_info:
        Gtin.parse("123")

    assert (
        str(exc_info.value)
        == "Failed to parse '123' as GTIN: Expected 8, 12, 13, or 14 digits, got 3."
    )


@pytest.mark.parametrize("value", ["0123456789abc", "0123456789⁰⁰⁰"])
def test_parse_nonnumeric_value(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        Gtin.parse(value)

    assert (
        str(exc_info.value)
        == f"Failed to parse {value!r} as GTIN: Expected a numerical value."
    )


def test_parse_gtin_13_with_invalid_check_digit() -> None:
    with pytest.raises(ParseError) as exc_info:
        Gtin.parse("5901234123458")

    assert (
        str(exc_info.value)
        == "Invalid GTIN check digit for '5901234123458': Expected 7, got 8."
    )


def test_parse_strips_surrounding_whitespace() -> None:
    gtin = Gtin.parse("  \t 5901234123457 \n  ")

    assert gtin.value == "5901234123457"


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-8
        "96385074",
        # 0-padded to GTIN-12
        "000096385074",
        # 0-padded to GTIN-13
        "0000096385074",
        # 0-padded to GTIN-14
        "00000096385074",
    ],
)
def test_parse_gtin_8(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_8,
        prefix=GS1Prefix(value="00009", usage="GS1 US"),
        company_prefix=GS1CompanyPrefix(value="0000963"),
        payload="9638507",
        check_digit=4,
    )


@pytest.mark.parametrize(
    "value",
    [
        "00000017",
        "00000123",
        "00001236",
        "00012348",
        "00123457",
        "01234565",
        "07038013",
    ],
)
def test_parse_gtin_8_with_leading_zeros(value: str) -> None:
    gtin = Gtin.parse(value)
    assert gtin.value == value
    assert gtin.format == GtinFormat.GTIN_8


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-12
        "614141000036",
        # 0-padded to GTIN-13
        "0614141000036",
        # 0-padded to GTIN-14
        "00614141000036",
    ],
)
def test_parse_gtin_12_without_leading_zero(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_12,
        prefix=GS1Prefix(value="061", usage="GS1 US"),
        company_prefix=GS1CompanyPrefix(value="0614141"),
        payload="61414100003",
        check_digit=6,
    )


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-12
        "036000291452",
        # 0-padded to GTIN-13
        "0036000291452",
        # 0-padded to GTIN-14
        "00036000291452",
    ],
)
def test_parse_gtin_12_with_1_leading_zero(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_12,
        prefix=GS1Prefix(value="003", usage="GS1 US"),
        company_prefix=GS1CompanyPrefix(value="0036000"),
        payload="03600029145",
        check_digit=2,
    )


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-12
        "006000291455",
        # 0-padded to GTIN-13
        "0006000291455",
        # 0-padded to GTIN-14
        "00006000291455",
    ],
)
def test_parse_gtin_12_with_2_leading_zero(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_12,
        prefix=GS1Prefix(value="0006", usage="GS1 US"),
        company_prefix=None,
        payload="00600029145",
        check_digit=5,
    )


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-12
        "000902914511",
        # 0-padded to GTIN-13
        "0000902914511",
        # 0-padded to GTIN-14
        "00000902914511",
    ],
)
def test_parse_gtin_12_with_3_leading_zero(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_12,
        prefix=GS1Prefix(value="00009", usage="GS1 US"),
        company_prefix=GS1CompanyPrefix(value="0000902"),
        payload="00090291451",
        check_digit=1,
    )


@pytest.mark.parametrize(
    "value",
    [
        # GTIN-13
        "5901234123457",
        # 0-padded to GTIN-14
        "05901234123457",
    ],
)
def test_parse_gtin_13(value: str) -> None:
    assert Gtin.parse(value) == Gtin(
        value=value,
        format=GtinFormat.GTIN_13,
        prefix=GS1Prefix(value="590", usage="GS1 Poland"),
        company_prefix=None,
        payload="590123412345",
        check_digit=7,
    )


def test_parse_gtin_14() -> None:
    assert Gtin.parse("98765432109213") == Gtin(
        value="98765432109213",
        format=GtinFormat.GTIN_14,
        prefix=GS1Prefix(value="876", usage="GS1 Netherlands"),
        company_prefix=None,
        payload="9876543210921",
        check_digit=3,
        packaging_level=9,
    )


def test_parse_gtin_with_unknown_gs1_prefix() -> None:
    assert Gtin.parse("6712670000276") == Gtin(
        value="6712670000276",
        format=GtinFormat.GTIN_13,
        prefix=None,
        company_prefix=None,
        payload="671267000027",
        check_digit=6,
    )



"""Support for GS1 Web URIs.

GS1 Web URIs are HTTP(S) URIs pointing to any hostname, optionally with a path
prefix, where GS1 element strings are encoded in the path and query parameters.

Examples of GS1 Web URIs:

- `https://id.gs1.org/gtin/614141123452/lot/ABC1/ser/12345?exp=180426`
- `https://id.gs1.org/gtin/614141123452?3103=000195`
- `https://id.gs1.org/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123`

This makes it possible to use GS1 Web URIs encoded in 2D barcodes both in
supply chain and logistics applications, as well as for consumers to look up
product information.

References:
    https://www.gs1.org/standards/Digital-Link/1-0

## Example

If you only want to parse GS1 Web URIs, you can import the GS1 Web URI parser
directly instead of using [`biip.parse()`][biip.parse].

    >>> from biip.gs1_web_uris import GS1WebURI

If the parsing succeeds, it returns a [`GS1WebURI`][biip.gs1_web_uris.GS1WebURI] object.

    >>> web_uri = GS1WebURI.parse("https://id.gs1.org/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123")

In this case, the URI is parsed into the SSCC of a shipping container, the GTIN
of the product within the shipping container, the number of item within, and the
batch number.

    >>> pprint(web_uri.element_strings)
    [
        GS1ElementString(
            ai=GS1ApplicationIdentifier(
                ai='00',
                description='Serial Shipping Container Code (SSCC)',
                data_title='SSCC',
                separator_required=False,
                format='N2+N18'
            ),
            value='106141412345678908',
            pattern_groups=[
                '106141412345678908'
            ],
            sscc=Sscc(
                value='106141412345678908',
                prefix=GS1Prefix(
                    value='061',
                    usage='GS1 US'
                ),
                company_prefix=GS1CompanyPrefix(
                    value='0614141'
                ),
                extension_digit=1,
                payload='10614141234567890',
                check_digit=8
            )
        ),
        GS1ElementString(
            ai=GS1ApplicationIdentifier(
                ai='02',
                description='Global Trade Item Number (GTIN) of contained trade items',
                data_title='CONTENT',
                separator_required=False,
                format='N2+N14'
            ),
            value='00614141123452',
            pattern_groups=[
                '00614141123452'
            ],
            gtin=Gtin(
                value='00614141123452',
                format=GtinFormat.GTIN_12,
                prefix=GS1Prefix(
                    value='061',
                    usage='GS1 US'
                ),
                company_prefix=GS1CompanyPrefix(
                    value='0614141'
                ),
                payload='61414112345',
                check_digit=2
            )
        ),
        GS1ElementString(
            ai=GS1ApplicationIdentifier(
                ai='37',
                description='Count of trade items or trade item pieces contained in a logistic unit',
                data_title='COUNT',
                separator_required=True,
                format='N2+N..8'
            ),
            value='25',
            pattern_groups=[
                '25'
            ]
        ),
        GS1ElementString(
            ai=GS1ApplicationIdentifier(
                ai='10',
                description='Batch or lot number',
                data_title='BATCH/LOT',
                separator_required=True,
                format='N2+X..20'
            ),
            value='ABC123',
            pattern_groups=[
                'ABC123'
            ]
        )
    ]
"""  # noqa: E501

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from urllib.parse import parse_qs, urlencode, urlparse, urlunsplit

from biip import ParseConfig, ParseError
from biip._exceptions import EncodeError
from biip.gs1_application_identifiers import (
    _GS1_APPLICATION_IDENTIFIERS,
    GS1ApplicationIdentifier,
)
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from biip.gs1_messages import GS1Message


@dataclass(frozen=True)
class GS1WebURI:
    """A GS1 Web URI is a URI that contains GS1 element strings."""

    value: str
    """Raw unprocessed value."""

    element_strings: GS1ElementStrings
    """List of element strings found in the message.

    See [`GS1ElementStrings`][biip.gs1_element_strings.GS1ElementStrings] for
    methods to extract interesting element strings from the list.
    """

    @classmethod
    def parse(  # noqa: C901
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,
    ) -> GS1WebURI:
        """Parse a string as a GS1 Web URI.

        Args:
            value: The string to parse.
            config: The parse configuration.

        Returns:
            The parsed GS1 Web URI.

        Raises:
            ParseError: If the parsing fails.
        """
        if config is None:
            config = ParseConfig()

        parsed = urlparse(value)

        match parsed.scheme:
            case "https" | "http":
                pass
            case "":
                msg = f"Expected URI, got {value!r}."
                raise ParseError(msg)
            case scheme:
                msg = f"Expected URI scheme to be 'http' or 'https', got {scheme!r}."
                raise ParseError(msg)

        # Consume path prefix
        path_parts = parsed.path.split("/")
        while path_parts:
            if path_parts[0] in _PRIMARY_IDENTIFIER_MAP:
                break
            path_parts.pop(0)

        if not path_parts:
            msg = f"Expected a primary identifier in the path, got '{parsed.path}'."
            raise ParseError(msg)

        if len(path_parts) % 2 != 0:
            path = f"/{'/'.join(path_parts)}"
            msg = f"Expected even number of path segments, got {path!r}."
            raise ParseError(msg)

        element_strings = GS1ElementStrings()
        path_pairs = _get_pairs(path_parts)

        # Extract exactly one primary identifier from path
        (pi_key, pi_value) = next(path_pairs)
        pi = _PRIMARY_IDENTIFIER_MAP[pi_key]
        if pi.zfill_to_width:
            pi_value = pi_value.zfill(pi.zfill_to_width)
        element_strings.append(
            GS1ElementString.extract(
                f"{pi.ai.ai}{pi_value}",
                config=config,
            )
        )

        # Extract qualifiers from path
        expected_qualifiers = list(pi.qualifiers)
        for qualifier_key, qualifier_value in path_pairs:
            qualifier = _get_qualifier(qualifier_key, expected_qualifiers)
            element_strings.append(
                GS1ElementString.extract(
                    f"{qualifier.ai.ai}{qualifier_value}",
                    config=config,
                )
            )

        # Extract query string components
        query_pairs = [
            (key, vals[-1])  # If an AI is repeated, the last value is used.
            for key, vals in parse_qs(parsed.query).items()
        ]
        for component_key, component_value in query_pairs:
            ai = _GS1_APPLICATION_IDENTIFIERS.get(component_key)
            if ai is None:
                # Extra query parameters are ignored.
                continue
            element_strings.append(
                GS1ElementString.extract(
                    f"{ai.ai}{component_value}",
                    config=config,
                )
            )

        return cls(value=value, element_strings=element_strings)

    @classmethod
    def from_element_strings(cls, element_strings: GS1ElementStrings) -> GS1WebURI:
        """Create a GS1 Web URI from a list of GS1 element strings.

        Args:
            element_strings: A list of GS1 element strings.

        Returns:
            GS1WebURI: The created GS1 Web URI.
        """
        return GS1WebURI(
            value=_build_url(element_strings),
            element_strings=element_strings,
        )

    def as_uri(
        self,
        *,
        domain: str | None = None,
        prefix: str | None = None,
        short_names: bool = False,
    ) -> str:
        """Render as a GS1 Web URI.

        Args:
            domain: The domain name to use in the URI. Defaults to `id.gs1.org`.
            prefix: The path prefix to use in the URI. Defaults to no prefix.
            short_names: Whether to use short names for AI values in the URI
                path. Defaults to False.

        Returns:
            str: The GS1 Web URI.
        """
        return _build_url(
            self.element_strings,
            domain=domain or "id.gs1.org",
            prefix=prefix,
            short_names=short_names,
        )

    def as_canonical_uri(self) -> str:
        """Render as a canonical GS1 Web URI.

        Canonical URIs:

        - Use the domain name `id.gs1.org`.
        - Uses numeric AI values in the URI path.
        - Excludes all query parameters that are not valid application identifiers.

        Returns:
            str: The canonical GS1 Web URI.

        References:
            GS1 Web URI Structure Standard, section 5.2
        """
        return _build_url(self.element_strings)

    def as_gs1_message(self) -> GS1Message:
        """Converts the GS1 Web URI to a GS1 Message."""
        from biip.gs1_messages import GS1Message

        return GS1Message.from_element_strings(self.element_strings)


def _get_pairs(iterable: Iterable[str]) -> Iterator[tuple[str, str]]:
    # Replace this function with itertools.batched(v, 2) when we require Python >= 3.12.
    iterator = iter(iterable)
    try:
        while True:
            yield next(iterator), next(iterator)
    except StopIteration:
        return


def _get_qualifier(
    qualifier_key: str,
    expected_qualifiers: list[_Component],
) -> _Component:
    if not expected_qualifiers:
        msg = f"Did not expect a qualifier, got {qualifier_key!r}."
        raise ParseError(msg)
    expected_qualifier_names = "/".join(
        f"{eq.ai.ai}/{eq.short_name}" for eq in expected_qualifiers
    )
    while expected_qualifiers:
        # Consume the expected qualifier, so that they can only be used
        # once, and only in order.
        qualifier = expected_qualifiers.pop(0)
        if qualifier_key in (qualifier.ai.ai, qualifier.short_name):
            return qualifier
    msg = (
        f"Expected one of {expected_qualifier_names} as qualifier, "
        f"got {qualifier_key!r}."
    )
    raise ParseError(msg)


def _build_url(
    element_strings: GS1ElementStrings,
    *,
    domain: str = "id.gs1.org",
    prefix: str | None = None,
    short_names: bool = False,
) -> str:
    primary_identifiers = [
        pi for pi in _PRIMARY_IDENTIFIERS if element_strings.get(ai=pi.ai)
    ]
    match primary_identifiers:
        case []:  # pragma: no cover  # Prevented by the parser.
            msg = "Expected exactly one primary identifier, none found."
            raise EncodeError(msg)
        case [pi]:
            primary_identifier = pi
        case _:  # pragma: no cover  # Prevented by the parser.
            msg = "Expected exactly one primary identifier, multiple found."
            raise EncodeError(msg)
    pi_element_string = element_strings.get(ai=primary_identifier.ai)
    assert pi_element_string

    qualifiers = {
        q: es
        for q in primary_identifier.qualifiers
        if (es := element_strings.get(ai=q.ai))
    }

    other_element_strings = [
        es
        for es in element_strings
        if es not in (pi_element_string, *qualifiers.values())
    ]

    if prefix is not None:
        if prefix.startswith("/"):
            msg = "Prefix must not start with '/'"
            raise ValueError(msg)
        if prefix.endswith("/"):
            msg = "Prefix must not end with '/'"
            raise ValueError(msg)
        path = prefix
    else:
        path = ""

    if short_names:
        path += f"/{primary_identifier.short_name}/{pi_element_string.value}"
    else:
        path += f"/{pi_element_string.ai.ai}/{pi_element_string.value}"

    for qualifier, element_string in qualifiers.items():
        if short_names:
            path += f"/{qualifier.short_name}/{element_string.value}"
        else:
            path += f"/{element_string.ai.ai}/{element_string.value}"

    params: dict[str, str] = {es.ai.ai: es.value for es in other_element_strings}

    return urlunsplit(
        [
            "https",
            domain,
            path,
            urlencode(params) if params else None,
            None,
        ]
    )


@dataclass(frozen=True)
class _Component:
    ai: GS1ApplicationIdentifier
    short_name: str
    zfill_to_width: int | None = None
    qualifiers: tuple[_Component, ...] = field(default_factory=tuple)


_PRIMARY_IDENTIFIERS = [
    _Component(
        ai=GS1ApplicationIdentifier.extract("01"),
        short_name="gtin",
        zfill_to_width=14,
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("22"),
                short_name="cpv",
            ),
            _Component(
                ai=GS1ApplicationIdentifier.extract("10"),
                short_name="lot",
            ),
            _Component(
                ai=GS1ApplicationIdentifier.extract("21"),
                short_name="ser",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8006"),
        short_name="itip",
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("22"),
                short_name="cpv",
            ),
            _Component(
                ai=GS1ApplicationIdentifier.extract("10"),
                short_name="lot",
            ),
            _Component(
                ai=GS1ApplicationIdentifier.extract("21"),
                short_name="ser",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8013"),
        short_name="gmn",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8010"),
        short_name="cpid",
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("8011"),
                short_name="cpsn",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("410"),
        short_name="shipTo",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("411"),
        short_name="billTo",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("412"),
        short_name="purchasedFrom",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("413"),
        short_name="shipFor",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("414"),
        short_name="gln",
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("254"),
                short_name="glnx",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("415"),
        short_name="payTo",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("416"),
        short_name="glnProd",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8017"),
        short_name="gsrnp",
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("8019"),
                short_name="srin",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8018"),
        short_name="gsrn",
        qualifiers=(
            _Component(
                ai=GS1ApplicationIdentifier.extract("8019"),
                short_name="srin",
            ),
        ),
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("255"),
        short_name="gcn",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("00"),
        short_name="sscc",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("253"),
        short_name="gdti",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("401"),
        short_name="ginc",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("402"),
        short_name="gsin",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8003"),
        short_name="grai",
    ),
    _Component(
        ai=GS1ApplicationIdentifier.extract("8004"),
        short_name="giai",
    ),
]
_PRIMARY_IDENTIFIER_MAP = {pi.short_name: pi for pi in _PRIMARY_IDENTIFIERS} | {
    pi.ai.ai: pi for pi in _PRIMARY_IDENTIFIERS
}



"""Support for Global Trade Item Number (GTIN).

The `biip.gtin` module contains Biip's support for parsing GTIN formats.

A GTIN is a number that uniquely identifies a trade item.

This class can interpet the following GTIN formats:

- GTIN-8, found in EAN-8 barcodes.
- GTIN-12, found in UPC-A and UPC-E barcodes.
- GTIN-13, found in EAN-13 barcodes.
- GTIN-14, found in ITF-14 barcodes, as well as a data field in GS1 barcodes.

If you only want to parse GTINs, you can import the GTIN parser directly
instead of using [`biip.parse()`][biip.parse].

    >>> from biip.gtin import Gtin

If parsing succeeds, it returns a [`Gtin`][biip.gtin.Gtin] object.

    >>> gtin = Gtin.parse("7032069804988")
    >>> pprint(gtin)
    Gtin(
        value='7032069804988',
        format=GtinFormat.GTIN_13,
        prefix=GS1Prefix(
            value='703',
            usage='GS1 Norway'
        ),
        company_prefix=GS1CompanyPrefix(
            value='703206'
        ),
        payload='703206980498',
        check_digit=8
    )

A GTIN can be converted to any other GTIN format, as long as the target
format is longer.

    >>> gtin.as_gtin_14()
    '07032069804988'

As all GTINs can be converted to GTIN-14 using
[`as_gtin_14()`][biip.gtin.Gtin.as_gtin_14], it is the recommended format to use
when storing or comparing GTINs. For example, when looking up a product
associated with a GTIN, the GTIN should first be expanded to a GTIN-14 before
using it to query the product catalog.
"""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Any

from biip import EncodeError, ParseError
from biip._parser import ParseConfig
from biip.checksums import gs1_standard_check_digit
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix

if TYPE_CHECKING:
    from collections.abc import Iterator


__all__ = ["Gtin", "GtinFormat"]


class GtinFormat(IntEnum):
    """Enum of GTIN formats."""

    GTIN_8 = 8
    """GTIN-8"""

    GTIN_12 = 12
    """GTIN-12"""

    GTIN_13 = 13
    """GTIN-13"""

    GTIN_14 = 14
    """GTIN-14"""

    def __str__(self) -> str:
        """Pretty string representation of format."""
        return self.name.replace("_", "-")

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"GtinFormat.{self.name}"

    @property
    def length(self) -> int:
        """Length of a GTIN of the given format."""
        return int(self)


@dataclass(frozen=True)
class Gtin:
    """Data class containing a GTIN."""

    value: str
    """Raw unprocessed value.

    May include leading zeros.
    """

    format: GtinFormat
    """[GTIN format][biip.gtin.GtinFormat], either GTIN-8, GTIN-12, GTIN-13, or
    GTIN-14.

    Classification is done after stripping leading zeros.
    """

    prefix: GS1Prefix | None
    """The [GS1 Prefix][biip.gs1_prefixes.GS1Prefix].

    Indicating what GS1 country organization that assigned
    code range.
    """

    company_prefix: GS1CompanyPrefix | None
    """The [GS1 Company Prefix][biip.gs1_prefixes.GS1CompanyPrefix].

    Identifying the company that issued the GTIN.
    """

    payload: str
    """The actual payload.

    Including packaging level if any, company prefix, and item reference.
    Excludes the check digit.
    """

    check_digit: int
    """Check digit used to check if the GTIN as a whole is valid."""

    packaging_level: int | None = None
    """Packaging level is the first digit in GTIN-14 codes.

    This digit is used for wholesale shipments, e.g. the GTIN-14 product
    identifier in GS1-128 barcodes, but not in the GTIN-13 barcodes used for
    retail products.
    """

    @classmethod
    def parse(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,
    ) -> Gtin:
        """Parse the given value into a [`Gtin`][biip.gtin.Gtin] object.

        Both GTIN-8, GTIN-12, GTIN-13, and GTIN-14 are supported.

        The checksum is guaranteed to be valid if a GTIN object is returned.

        Args:
            value: The value to parse.
            config: Configuration options for parsing.

        Returns:
            GTIN data structure with the successfully extracted data.

        Raises:
            ParseError: If the parsing fails.
        """
        if config is None:
            config = ParseConfig()

        from biip.rcn import Rcn

        value = value.strip()

        if len(value) not in (8, 12, 13, 14):
            msg = (
                f"Failed to parse {value!r} as GTIN: "
                f"Expected 8, 12, 13, or 14 digits, got {len(value)}."
            )
            raise ParseError(msg)

        if not value.isdecimal():
            msg = f"Failed to parse {value!r} as GTIN: Expected a numerical value."
            raise ParseError(msg)

        stripped_value = _strip_leading_zeros(value)
        assert len(stripped_value) in (8, 12, 13, 14)

        num_significant_digits = len(stripped_value)
        gtin_format = GtinFormat(num_significant_digits)

        payload = stripped_value[:-1]
        check_digit = int(stripped_value[-1])

        packaging_level: int | None = None
        prefix_value = stripped_value
        if gtin_format == GtinFormat.GTIN_14:
            packaging_level = int(stripped_value[0])
            prefix_value = stripped_value[1:]
        elif gtin_format == GtinFormat.GTIN_12:
            # Add a zero to convert U.P.C. Company Prefix to GS1 Company Prefix
            prefix_value = stripped_value.zfill(13)
        elif gtin_format == GtinFormat.GTIN_8:
            prefix_value = stripped_value.zfill(12)

        prefix = GS1Prefix.extract(prefix_value)
        company_prefix = GS1CompanyPrefix.extract(prefix_value)

        calculated_check_digit = gs1_standard_check_digit(payload)
        if check_digit != calculated_check_digit:
            msg = (
                f"Invalid GTIN check digit for {value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )
            raise ParseError(msg)

        gtin_type: type[Gtin | Rcn]
        if (
            gtin_format <= GtinFormat.GTIN_13
            and prefix is not None
            and "Restricted Circulation Number" in prefix.usage
        ):
            gtin_type = Rcn
        else:
            gtin_type = Gtin

        result = gtin_type(
            value=value,
            format=gtin_format,
            prefix=prefix,
            company_prefix=company_prefix,
            payload=payload,
            check_digit=check_digit,
            packaging_level=packaging_level,
        )

        if isinstance(result, Rcn):
            result = result._with_usage()  # noqa: SLF001
            result = result._parsed_with_regional_rules(  # noqa: SLF001
                config=config
            )

        return result

    def __rich_repr__(self) -> Iterator[tuple[str, Any] | tuple[str, Any, Any]]:  # noqa: D105
        # Skip printing fields with default values
        yield "value", self.value
        yield "format", self.format
        yield "prefix", self.prefix
        yield "company_prefix", self.company_prefix
        yield "payload", self.payload
        yield "check_digit", self.check_digit
        yield "packaging_level", self.packaging_level, None

    def as_gtin_8(self) -> str:
        """Format as a GTIN-8."""
        return self._as_format(GtinFormat.GTIN_8)

    def as_gtin_12(self) -> str:
        """Format as a GTIN-12."""
        return self._as_format(GtinFormat.GTIN_12)

    def as_gtin_13(self) -> str:
        """Format as a GTIN-13."""
        return self._as_format(GtinFormat.GTIN_13)

    def as_gtin_14(self) -> str:
        """Format as a GTIN-14."""
        return self._as_format(GtinFormat.GTIN_14)

    def _as_format(self, gtin_format: GtinFormat) -> str:
        if self.format.length > gtin_format.length:
            msg = f"Failed encoding {self.value!r} as {gtin_format!s}."
            raise EncodeError(msg)
        return f"{self.payload}{self.check_digit}".zfill(gtin_format.length)

    def without_variable_measure(self) -> Gtin:
        """Create a new GTIN where the variable measure is zeroed out.

        This method is a no-op for proper GTINs. For RCNs, see the method on the
        `Rcn` subclass.

        Returns:
            A GTIN instance with zeros in the variable measure places.

        Raises:
            EncodeError: If the rules for variable measures in the region are unknown.
        """
        return self


def _strip_leading_zeros(value: str) -> str:
    if len(value) in (12, 13, 14) and len(value.lstrip("0")) in (9, 10, 11, 12):
        # Keep up to three leading zeros in GTIN-12
        num_zeros_before_gtin_12 = len(value) - 12
        return value[num_zeros_before_gtin_12:]

    if len(value) >= 8 and len(value.lstrip("0")) <= 8:
        # Keep all leading zeros in GTIN-8
        num_zeros_before_gtin_8 = len(value) - 8
        return value[num_zeros_before_gtin_8:]

    return value.lstrip("0")



"""Support for barcode data with GS1 messages.

The `biip.gs1_messages` module contains Biip's support for parsing data
consisting of GS1 messages and element strings. GS1 messages are text strings
consisting of one or more GS1 element strings. Each GS1 Element String is
identified by a GS1 Application Identifier (AI) prefix followed by the element's
value.

Data of this format is found in the following types of barcodes:

- [GS1-128](https://en.wikipedia.org/wiki/Code_128)
- [GS1 DataBar](https://en.wikipedia.org/wiki/GS1_DataBar_Coupon)
- [GS1 DataMatrix](https://en.wikipedia.org/wiki/Data_Matrix)
- [GS1 QR Code](https://en.wikipedia.org/wiki/QR_code)

If you only want to parse GS1 Messages, you can import the GS1 Message parser
directly instead of using [`biip.parse()`][biip.parse].

    >>> from biip.gs1_messages import GS1Message

If the parsing succeeds, it returns a [`GS1Message`][biip.gs1_messages.GS1Message]
object.

    >>> msg = GS1Message.parse("010703206980498815210526100329")

The [`GS1Message`][biip.gs1_messages.GS1Message] can be represented as an HRI,
short for "human readable interpretation", using
[`msg.as_hri()`][biip.gs1_messages.GS1Message.as_hri]. The HRI is the text
usually printed below or next to the barcode.

    >>> msg.value
    '010703206980498815210526100329'
    >>> msg.as_hri()
    '(01)07032069804988(15)210526(10)0329'

HRI can also be parsed using
[`GS1Message.parse_hri()`][biip.gs1_messages.GS1Message.parse_hri].

    >>> msg = GS1Message.parse_hri("(01)07032069804988(15)210526(10)0329")

A message can contain multiple element strings.

    >>> len(msg.element_strings)
    3

In this example, the first element string is a GTIN.

    >>> pprint(msg.element_strings[0])
    GS1ElementString(
        ai=GS1ApplicationIdentifier(
            ai='01',
            description='Global Trade Item Number (GTIN)',
            data_title='GTIN',
            separator_required=False,
            format='N2+N14'
        ),
        value='07032069804988',
        pattern_groups=[
            '07032069804988'
        ],
        gtin=Gtin(
            value='07032069804988',
            format=GtinFormat.GTIN_13,
            prefix=GS1Prefix(
                value='703',
                usage='GS1 Norway'
            ),
            company_prefix=GS1CompanyPrefix(
                value='703206'
            ),
            payload='703206980498',
            check_digit=8
        )
    )

The `element_strings` attribute has
[`element_strings.get()`][biip.gs1_element_strings.GS1ElementStrings.get] and
[`element_strings.filter()`][biip.gs1_element_strings.GS1ElementStrings.filter]
methods to lookup element strings either by the Application Identifier's "data
title" or its AI number.

    >>> pprint(msg.element_strings.get(data_title='BEST BY'))
    GS1ElementString(
        ai=GS1ApplicationIdentifier(
            ai='15',
            description='Best before date (YYMMDD)',
            data_title='BEST BEFORE or BEST BY',
            separator_required=False,
            format='N2+N6'
        ),
        value='210526',
        pattern_groups=[
            '210526'
        ],
        date=datetime.date(2021, 5, 26)
    )
    >>> pprint(msg.element_strings.get(ai="10"))
    GS1ElementString(
        ai=GS1ApplicationIdentifier(
            ai='10',
            description='Batch or lot number',
            data_title='BATCH/LOT',
            separator_required=True,
            format='N2+X..20'
        ),
        value='0329',
        pattern_groups=[
            '0329'
        ]
    )
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING

from biip import ParseError
from biip._parser import ParseConfig
from biip.gs1_application_identifiers import _GS1_APPLICATION_IDENTIFIERS
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings

if TYPE_CHECKING:
    from biip.gs1_web_uris import GS1WebURI


@dataclass(frozen=True)
class GS1Message:
    """A GS1 message is the result of a single barcode scan.

    It may contain one or more GS1 element strings.

    Examples:
        See `biip.gs1_messages` for a usage example.
    """

    value: str
    """Raw unprocessed value."""

    element_strings: GS1ElementStrings
    """List of element strings found in the message.

    See [`GS1ElementStrings`][biip.gs1_element_strings.GS1ElementStrings] for
    methods to extract interesting element strings from the list.
    """

    @classmethod
    def parse(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,
    ) -> GS1Message:
        """Parse a string from a barcode scan as a GS1 message with AIs.

        Args:
            value: The string to parse.
            config: Configuration options for parsing.

        Returns:
            A message object with one or more element strings.

        Raises:
            ValueError: If the `separator_char` isn't exactly 1 character long.
            ParseError: If the parsing fails.
        """
        if config is None:
            config = ParseConfig()

        value = value.strip()
        element_strings = GS1ElementStrings()
        rest = value[:]

        while rest:
            element_string = GS1ElementString.extract(rest, config=config)
            element_strings.append(element_string)

            rest = rest[len(element_string) :]

            # Separator characters are accepted inbetween any element string,
            # even if the AI doesn't require it. See GS1 General Specifications,
            # section 7.8.6 for details.
            while rest.startswith(tuple(config.separator_chars)):
                rest = rest[1:]

        return cls(value=value, element_strings=element_strings)

    @classmethod
    def from_element_strings(cls, element_strings: GS1ElementStrings) -> GS1Message:
        """Create a GS1 message from a list of element strings.

        Args:
            element_strings: A list of GS1 element strings.

        Returns:
            GS1Message: The created GS1 message.
        """
        parts = chain(
            *[
                [
                    es.ai.ai,
                    es.value,
                    ("\x1d" if es.ai.separator_required else ""),
                ]
                for es in element_strings
            ]
        )
        normalized_string = "".join(parts).removesuffix("\x1d")
        return cls(value=normalized_string, element_strings=element_strings)

    @classmethod
    def parse_hri(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,
    ) -> GS1Message:
        """Parse the GS1 string given in HRI (human readable interpretation) format.

        Args:
            value: The HRI string to parse.
            config: Configuration options for parsing.

        Returns:
            A message object with one or more element strings.

        Raises:
            ParseError: If parsing of the data fails.
        """
        if config is None:
            config = ParseConfig()

        value = value.strip()
        if not value.startswith("("):
            msg = f"Expected HRI string {value!r} to start with a parenthesis."
            raise ParseError(msg)

        pattern = r"\((\d+)\)(\w+)"
        matches: list[tuple[str, str]] = re.findall(pattern, value)
        if not matches:
            msg = (
                f"Could not find any GS1 Application Identifiers in {value!r}. "
                "Expected format: '(AI)DATA(AI)DATA'."
            )
            raise ParseError(msg)

        element_strings = GS1ElementStrings()
        for ai_number, ai_data in matches:
            if ai_number not in _GS1_APPLICATION_IDENTIFIERS:
                msg = f"Unknown GS1 Application Identifier {ai_number!r} in {value!r}."
                raise ParseError(msg)
            element_strings.append(
                GS1ElementString.extract(f"{ai_number}{ai_data}", config=config)
            )

        return GS1Message.from_element_strings(element_strings)

    def as_hri(self) -> str:
        """Render as a human readable interpretation (HRI).

        The HRI is often printed directly below barcodes.

        Returns:
            A human-readable string where the AIs are wrapped in parenthesis.
        """
        return "".join(es.as_hri() for es in self.element_strings)

    def as_gs1_web_uri(self) -> GS1WebURI:
        """Convert to a GS1 Web URI."""
        from biip.gs1_web_uris import GS1WebURI

        return GS1WebURI.from_element_strings(self.element_strings)



"""Tests of parsing GLNs."""

import pytest

from biip import ParseError
from biip.gln import Gln
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix


def test_parse() -> None:
    gln = Gln.parse("1234567890128")

    assert gln == Gln(
        value="1234567890128",
        prefix=GS1Prefix(value="123", usage="GS1 US"),
        company_prefix=GS1CompanyPrefix(value="1234567890"),
        payload="123456789012",
        check_digit=8,
    )


def test_parse_strips_surrounding_whitespace() -> None:
    gln = Gln.parse("  \t 1234567890128 \n  ")

    assert gln.value == "1234567890128"


def test_parse_value_with_invalid_length() -> None:
    with pytest.raises(ParseError) as exc_info:
        Gln.parse("123")

    assert (
        str(exc_info.value)
        == "Failed to parse '123' as GLN: Expected 13 digits, got 3."
    )


@pytest.mark.parametrize("value", ["123456789o128", "123456789⁰128"])
def test_parse_nonnumeric_value(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        Gln.parse(value)

    assert (
        str(exc_info.value)
        == f"Failed to parse {value!r} as GLN: Expected a numerical value."
    )


def test_parse_with_invalid_check_digit() -> None:
    with pytest.raises(ParseError) as exc_info:
        Gln.parse("1234567890127")

    assert (
        str(exc_info.value)
        == "Invalid GLN check digit for '1234567890127': Expected 8, got 7."
    )


def test_as_gln() -> None:
    gln = Gln.parse("  \t 1234567890128 \n  ")

    assert gln.as_gln() == "1234567890128"



"""Script to download and extract Application Identifier data from GS1."""

import dataclasses
import json

import httpx

from biip.gs1_application_identifiers import GS1ApplicationIdentifier

AI_URL = "https://ref.gs1.org/ai/GS1_Application_Identifiers.jsonld"


def main() -> None:
    """The script's main function."""
    html_content = download(AI_URL)
    application_identifiers = parse(html_content)
    output(application_identifiers)


def download(url: str) -> bytes:
    """Download the data from GS1."""
    return httpx.get(url, timeout=30).content


def parse(json_content: bytes) -> list[GS1ApplicationIdentifier]:
    """Parse the data from HTML to GS1ApplicationIdentifier objects."""
    result: list[GS1ApplicationIdentifier] = []

    data = json.loads(json_content)

    for row in data["applicationIdentifiers"]:
        if "applicationIdentifier" not in row:
            continue
        result.append(
            GS1ApplicationIdentifier(
                ai=row["applicationIdentifier"],
                description=row["description"],
                format=row["formatString"],
                data_title=row["title"],
                separator_required=row["separatorRequired"],
                pattern=rf"^{row['applicationIdentifier']}{_fix_pattern(row['regex'])}$",
            )
        )

    return result


def _fix_pattern(value: str) -> str:
    """Fix errors in regex patterns."""
    # Add missing opening square bracket to the regex for AI 703x
    if value == r"(\d{3})(!%-?A-Z_a-z\x22]{1,27})":
        return r"(\d{3})([!%-?A-Z_a-z\x22]{1,27})"
    # Add missing opening square bracket to the regex for AI 723x
    if value == r"(!%-?A-Z_a-z\x22]{3,30})":
        return r"([!%-?A-Z_a-z\x22]{3,30})"
    return value


def output(application_identifiers: list[GS1ApplicationIdentifier]) -> None:
    """Output the GS1ApplicationIdentifier objects as JSON to stdout."""
    print(
        json.dumps([dataclasses.asdict(ai) for ai in application_identifiers], indent=2)
    )


if __name__ == "__main__":
    main()



"""Script to download and extract Company Prefix data from GS1."""

import json
from typing import Union
from xml.etree import ElementTree  # noqa: ICN001

import httpx

TrieNode = Union[dict[str, "TrieNode"], int]  # noqa: UP007

COMPANY_PREFIX_URL = "https://www.gs1.org/docs/gcp_length/gcpprefixformatlist.xml"


def main() -> None:
    """The script's main function."""
    xml_content = download(COMPANY_PREFIX_URL)
    trie = parse(xml_content)
    output(trie)


def download(url: str) -> bytes:
    """Download the data from GS1."""
    return httpx.get(url, timeout=30).content


def parse(xml_content: bytes) -> TrieNode:
    """Parse the data from XML to a trie to make prefix length lookup easy."""
    trie: TrieNode = {}

    doc = ElementTree.fromstring(xml_content)  # noqa: S314
    for entry in doc.findall("entry"):
        shared_prefix = entry.attrib["prefix"]
        company_prefix_length = int(entry.attrib["gcpLength"])

        # Build/traverse the trie
        node: TrieNode = trie
        digits = list(shared_prefix)
        while digits:
            assert isinstance(node, dict)
            digit = digits.pop(0)
            if digits:
                if digit not in node:
                    node[digit] = {}
                node = node[digit]
            else:
                node[digit] = company_prefix_length

    return trie


def output(trie: TrieNode) -> None:
    """Output the trie data structure as JSON to stdout."""
    print(json.dumps(trie, indent=2))


if __name__ == "__main__":
    main()



"""Test of GS1 prefixes."""

import pytest

from biip import ParseError
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix


@pytest.mark.parametrize("bad_value", ["abcdef", "1a2b3c"])
def test_invalid_gs1_prefix(bad_value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1Prefix.extract(bad_value)

    assert str(exc_info.value) == f"Failed to get GS1 Prefix from {bad_value!r}."


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            # Undefined prefix
            "0",
            None,
        ),
        (
            # In the GTIN-8 range
            "0000001999",
            GS1Prefix(value="0000001", usage="Unused to avoid collision with GTIN-8"),
        ),
        (
            # In the GTIN-12 range
            "060999",
            GS1Prefix(value="060", usage="GS1 US"),
        ),
        (
            # In the GTIN-13 range
            "139999",
            GS1Prefix(value="139", usage="GS1 US"),
        ),
        (
            # Unassigned prefix
            "6712670000276",
            None,
        ),
        (
            # Norwegian GTIN-13
            "701999",
            GS1Prefix(value="701", usage="GS1 Norway"),
        ),
        (
            # ISBN
            "9781492053743",
            GS1Prefix(value="978", usage="Bookland (ISBN)"),
        ),
        #
        # Exceptions defined in the GS1 prefix table's footnotes
        ("960", GS1Prefix(value="960", usage="GS1 UK - GTIN-8")),
        ("961", GS1Prefix(value="961", usage="GS1 UK - GTIN-8")),
        ("9620", GS1Prefix(value="9620", usage="GS1 UK - GTIN-8")),
        ("9621", GS1Prefix(value="9621", usage="GS1 UK - GTIN-8")),
        ("9622", GS1Prefix(value="9622", usage="GS1 UK - GTIN-8")),
        ("9623", GS1Prefix(value="9623", usage="GS1 UK - GTIN-8")),
        ("9624", GS1Prefix(value="9624", usage="GS1 UK - GTIN-8")),
        ("9625", GS1Prefix(value="9625", usage="GS1 Poland - GTIN-8")),
        ("9626", GS1Prefix(value="9626", usage="GS1 Poland - GTIN-8")),
        ("9627", GS1Prefix(value="9627", usage="GS1 Global Office - GTIN-8")),
        ("9628", GS1Prefix(value="9628", usage="GS1 Global Office - GTIN-8")),
        ("9629", GS1Prefix(value="9629", usage="GS1 Global Office - GTIN-8")),
    ],
)
def test_gs1_prefix(value: str, expected: GS1Prefix) -> None:
    assert GS1Prefix.extract(value) == expected


def test_gs1_prefix_is_hashable() -> None:
    prefix = GS1Prefix.extract("978")

    assert hash(prefix) is not None


@pytest.mark.parametrize("bad_value", ["abcdef", "1a2b3c"])
def test_invalid_gs1_company_prefix(bad_value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1CompanyPrefix.extract(bad_value)

    assert (
        str(exc_info.value) == f"Failed to get GS1 Company Prefix from {bad_value!r}."
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            # Undefined prefix
            "0",
            None,
        ),
        (
            # Undefined prefix
            "0000031000000",
            None,
        ),
        # GLNs
        (
            # Oda's fulfillment center in Helsinki
            "6429830062707",
            GS1CompanyPrefix(value="64298300627"),
        ),
        # GTINs
        (
            # Soda from Ringnes AS
            "7044610873466",
            GS1CompanyPrefix(value="704461"),
        ),
        (
            # Chocolate from Ferrero SpA
            "8000500247167",
            GS1CompanyPrefix(value="8000500"),
        ),
        (
            # ISBNs are defined to have a prefix length of 0.
            "9781492053743",
            None,
        ),
        # SSCCs
        (
            # Coffee delivery from Nestlé
            "376130321109103420",
            GS1CompanyPrefix(value="376130321"),
        ),
    ],
)
def test_gs1_company_prefix(value: str, expected: GS1CompanyPrefix) -> None:
    assert GS1CompanyPrefix.extract(value) == expected


def test_gs1_company_prefix_is_hashable() -> None:
    prefix = GS1CompanyPrefix.extract("7044610873466")

    assert hash(prefix) is not None



import pytest

from biip.checksums import gs1_price_weight_check_digit, gs1_standard_check_digit


@pytest.mark.parametrize("value", ["abc", "⁰⁰⁰"])
def test_gs1_standard_check_digit_with_nonnumeric_value(value: str) -> None:
    with pytest.raises(
        ValueError,
        match=rf"^Expected numeric value, got {value!r}.$",
    ):
        gs1_standard_check_digit(value)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # Example from reference
        ("37610425002123456", 9),
        # GTIN-13 from a product
        ("629104150021", 3),
        ("703801005447", 1),
        ("950110153100", 0),
        # ISBN from a book
        ("978820551237", 5),
        # GLN for an organization
        ("708000382434", 9),
    ],
)
def test_gs1_standard_check_digit(value: str, expected: int) -> None:
    assert gs1_standard_check_digit(value) == expected


@pytest.mark.parametrize("value", ["abc", "⁰⁰⁰"])
def test_gs1_price_weight_check_digit_with_nonnumeric_value(value: str) -> None:
    with pytest.raises(
        ValueError,
        match=rf"^Expected numeric value, got {value!r}.$",
    ):
        gs1_price_weight_check_digit(value)


@pytest.mark.parametrize(
    "value",
    [
        # Too short
        "123",
        # Too long
        "123456",
    ],
)
def test_gs1_price_weight_check_digit_on_values_with_wrong_length(value: str) -> None:
    with pytest.raises(
        ValueError,
        match=rf"^Expected input of length 4 or 5, got {value!r}.$",
    ):
        gs1_price_weight_check_digit(value)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # Four-digit price field
        ("2875", 9),
        # Five-digit price field
        ("14685", 6),
    ],
)
def test_gs1_price_weight_check_digit(value: str, expected: int) -> None:
    assert gs1_price_weight_check_digit(value) == expected



"""GS1 element strings.

An element string consists of a GS1 Application Identifier (AI) and its data field.

Element strings are usually found in a GS1 message or a GS1 Web URI. A single
barcode can contain multiple element strings.

Examples:
    >>> from biip.gs1_element_strings import GS1ElementString
    >>> element_string = GS1ElementString.extract("0107032069804988")
    >>> pprint(element_string)
    GS1ElementString(
        ai=GS1ApplicationIdentifier(
            ai='01',
            description='Global Trade Item Number (GTIN)',
            data_title='GTIN',
            separator_required=False,
            format='N2+N14'
        ),
        value='07032069804988',
        pattern_groups=[
            '07032069804988'
        ],
        gtin=Gtin(
            value='07032069804988',
            format=GtinFormat.GTIN_13,
            prefix=GS1Prefix(
                value='703',
                usage='GS1 Norway'
            ),
            company_prefix=GS1CompanyPrefix(
                value='703206'
            ),
            payload='703206980498',
            check_digit=8
        )
    )
    >>> element_string.as_hri()
    '(01)07032069804988'
"""

from __future__ import annotations

import calendar
import datetime as dt
import re
from collections.abc import Iterator
from dataclasses import dataclass, replace
from decimal import Decimal
from typing import TYPE_CHECKING, Any

from biip import ParseError
from biip._parser import ParseConfig
from biip.gln import Gln
from biip.gs1_application_identifiers import GS1ApplicationIdentifier
from biip.gtin import Gtin
from biip.sscc import Sscc

if TYPE_CHECKING:
    from collections.abc import Iterator


try:
    import moneyed  # noqa: TC002

    have_moneyed = True
except ImportError:  # pragma: no cover
    have_moneyed = False


@dataclass(frozen=True)
class GS1ElementString:
    """A GS1 element string consists of an Application Identifier (AI) and a value."""

    ai: GS1ApplicationIdentifier
    """The element's Application Identifier (AI)."""

    value: str
    """Raw data field of the Element String. Does not include the AI."""

    pattern_groups: list[str]
    """List of pattern groups extracted from the Element String."""

    gln: Gln | None = None
    """A GLN created from the element string, if the AI represents a GLN."""

    gln_error: str | None = None
    """The GLN parse error, if parsing as a GLN was attempted and failed."""

    gtin: Gtin | None = None
    """A GTIN created from the element string, if the AI represents a GTIN."""

    gtin_error: str | None = None
    """The GTIN parse error, if parsing as a GTIN was attempted and failed."""

    sscc: Sscc | None = None
    """An SSCC created from the element string, if the AI represents a SSCC."""

    sscc_error: str | None = None
    """The SSCC parse error, if parsing as an SSCC was attempted and failed."""

    date: dt.date | None = None
    """A date created from the element string, if the AI represents a date."""

    datetime: dt.datetime | None = None
    """A datetime created from the element string, if the AI represents a datetime."""

    decimal: Decimal | None = None
    """A decimal value created from the element string, if the AI represents a
    number."""

    money: moneyed.Money | None = None
    """A Money value created from the element string, if the AI represents a
    currency and an amount.

    Only set if `py-moneyed` is installed.
    """

    @classmethod
    def extract(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,
    ) -> GS1ElementString:
        """Extract the first GS1 Element String from the given value.

        If the element string contains a primitive data type, like a date,
        decimal number, or currency, it will be parsed and stored in the
        [`date`][biip.gs1_messages.GS1ElementString.date],
        [`decimal`][biip.gs1_messages.GS1ElementString.decimal], or
        [`money`][biip.gs1_messages.GS1ElementString.money] field respectively.
        If parsing of a primitive data type fails, a
        [`ParseError`][biip.ParseError] will be raised.

        If the element string contains another supported format, like a GLN,
        GTIN, or SSCC, it will parsed and validated, and the result stored in
        the fields
        [`gln`][biip.gs1_messages.GS1ElementString.gln],
        [`gtin`][biip.gs1_messages.GS1ElementString.gtin], or
        [`sscc`][biip.gs1_messages.GS1ElementString.sscc] respectively. If parsing or
        validation of an inner format fails, the
        [`gln_error`][biip.gs1_messages.GS1ElementString.gln_error],
        [`gtin_error`][biip.gs1_messages.GS1ElementString.gtin_error], or
        [`sscc_error`][biip.gs1_messages.GS1ElementString.sscc_error] field will be set.
        No [`ParseError`][biip.ParseError] will be raised.

        Args:
            value: The string to extract an Element String from. May contain
                more than one Element String.
            config: Configuration options for parsing.

        Returns:
            A data class with the Element String's parts and data extracted from it.

        Raises:
            ValueError: If the `separator_char` isn't exactly 1 character long.
            ParseError: If the parsing fails.
        """
        if config is None:
            config = ParseConfig()

        if any(len(char) != 1 for char in config.separator_chars):
            msg = (
                "All separator characters must be exactly 1 character long, "
                f"got {list(config.separator_chars)!r}."
            )
            raise ValueError(msg)

        ai = GS1ApplicationIdentifier.extract(value)

        for separator_char in config.separator_chars:
            value = value.split(separator_char, maxsplit=1)[0]

        pattern = ai.pattern.removesuffix("$")
        matches = re.match(pattern, value)
        if not matches:
            msg = f"Failed to match {value!r} with GS1 AI {ai} pattern '{ai.pattern}'."
            raise ParseError(msg)
        pattern_groups = [group for group in matches.groups() if group is not None]
        value = "".join(pattern_groups)

        element = cls(ai=ai, value=value, pattern_groups=pattern_groups)
        element = element._with_gln(config=config)
        element = element._with_gtin(config=config)  # noqa: SLF001
        element = element._with_sscc(config=config)  # noqa: SLF001
        element = element._with_date_and_datetime(config=config)  # noqa: SLF001
        element = element._with_decimal()  # noqa: SLF001

        return element  # noqa: RET504

    def _with_gln(self, *, config: ParseConfig) -> GS1ElementString:
        if self.ai.ai[:2] != "41":
            return self

        try:
            gln = Gln.parse(self.value, config=config)
            gln_error = None
        except ParseError as exc:
            gln = None
            gln_error = str(exc)

        return replace(self, gln=gln, gln_error=gln_error)

    def _with_gtin(self, *, config: ParseConfig) -> GS1ElementString:
        if self.ai.ai not in ("01", "02", "03"):
            return self

        try:
            gtin = Gtin.parse(self.value, config=config)
            gtin_error = None
        except ParseError as exc:
            gtin = None
            gtin_error = str(exc)

        return replace(self, gtin=gtin, gtin_error=gtin_error)

    def _with_sscc(self, *, config: ParseConfig) -> GS1ElementString:
        if self.ai.ai != "00":
            return self

        try:
            sscc = Sscc.parse(self.value, config=config)
            sscc_error = None
        except ParseError as exc:
            sscc = None
            sscc_error = str(exc)

        return replace(self, sscc=sscc, sscc_error=sscc_error)

    def _with_date_and_datetime(self, *, config: ParseConfig) -> GS1ElementString:
        if self.ai.ai not in (
            "11",
            "12",
            "13",
            "15",
            "16",
            "17",
            "4324",
            "4325",
            "4326",
            "7003",
            "7006",
            "7007",
            "7011",
            "8008",
        ):
            return self

        try:
            date, datetime = _parse_date_and_datetime(self.value)
        except ValueError as exc:
            if not config.gs1_element_strings_verify_date:
                return self
            msg = f"Failed to parse GS1 AI {self.ai} date/time from {self.value!r}."
            raise ParseError(msg) from exc

        return replace(self, date=date, datetime=datetime)

    def _with_decimal(self) -> GS1ElementString:
        variable_measure = self.ai.ai[:2] in (
            "31",
            "32",
            "33",
            "34",
            "35",
            "36",
        )
        amount_payable = self.ai.ai[:3] in ("390", "392")
        amount_payable_with_currency = self.ai.ai[:3] in ("391", "393")
        percentage = self.ai.ai[:3] in ("394",)

        if (
            variable_measure
            or amount_payable
            or amount_payable_with_currency
            or percentage
        ):
            # See GS1 General Specifications, chapter 3.6 for details.

            # Only group for variable_measure, amount_payable, and percentage.
            # Second and last group for amount_payable_with_currency.
            value = self.pattern_groups[-1]

            num_decimals = int(self.ai.ai[3])
            num_units = len(value) - num_decimals

            units = value[:num_units]
            decimals = value[num_units:]

            decimal = Decimal(f"{units}.{decimals}")
        else:
            decimal = None

        if amount_payable_with_currency and have_moneyed:
            import moneyed

            currency = moneyed.get_currency(iso=self.pattern_groups[0])
            money = moneyed.Money(amount=decimal, currency=currency)
        else:
            money = None

        return replace(self, decimal=decimal, money=money)

    def __len__(self) -> int:
        """Get the length of the element string."""
        return len(self.ai) + len(self.value)

    def __rich_repr__(self) -> Iterator[tuple[str, Any] | tuple[str, Any, Any]]:  # noqa: D105
        # Skip printing fields with default values
        yield "ai", self.ai
        yield "value", self.value
        yield "pattern_groups", self.pattern_groups
        yield "gln", self.gln, None
        yield "gln_error", self.gln_error, None
        yield "gtin", self.gtin, None
        yield "gtin_error", self.gtin_error, None
        yield "sscc", self.sscc, None
        yield "sscc_error", self.sscc_error, None
        yield "date", self.date, None
        yield "datetime", self.datetime, None
        yield "decimal", self.decimal, None
        yield "money", self.money, None

    def as_hri(self) -> str:
        """Render as a human readable interpretation (HRI).

        The HRI is often printed directly below the barcode.

        Returns:
            A human-readable string where the AI is wrapped in parenthesis.
        """
        return f"{self.ai}{self.value}"


def _parse_date_and_datetime(value: str) -> tuple[dt.date, dt.datetime | None]:
    pairs = [value[i : i + 2] for i in range(0, len(value), 2)]

    year = int(pairs[0])
    year += _get_century(year)
    month = int(pairs[1])
    day = int(pairs[2])
    if day == 0:
        day = _get_last_day_of_month(year, month)
    date = dt.date(year, month, day)
    if not pairs[3:]:
        return date, None

    hour = int(pairs[3])
    minute = int(pairs[4] if pairs[4:] else 0)
    seconds = int(pairs[5] if pairs[5:] else 0)
    if hour == 99 and minute == 99:
        return date, None

    return date, dt.datetime(year, month, day, hour, minute, seconds)  # noqa: DTZ001


def _get_century(two_digit_year: int) -> int:
    """Get century from two-digit year.

    The two-digit year refers to a year in the range between 49 years past
    and 50 years into the future.

    Args:
        two_digit_year: A two-digit year, e.g. without century specified.

    Returns:
        The century the year is in.

    References:
        GS1 General Specifications, section 7.12
    """
    current_year = dt.datetime.now(tz=dt.timezone.utc).year
    current_century = current_year - current_year % 100
    two_digit_current_year = current_year - current_century

    # Previous century
    if 51 <= two_digit_year - two_digit_current_year <= 99:
        return current_century - 100

    # Next century
    if -99 <= two_digit_year - two_digit_current_year <= -50:
        # Skipping coverage as this code won't run until year 2051
        return current_century + 100  # pragma: no cover

    # Current century
    return current_century


def _get_last_day_of_month(year: int, month: int) -> int:
    """Get the last day of the given month."""
    return calendar.monthrange(year, month)[1]


class GS1ElementStrings(list[GS1ElementString]):
    """List of GS1 element strings."""

    def filter(
        self,
        *,
        ai: str | GS1ApplicationIdentifier | None = None,
        data_title: str | None = None,
    ) -> GS1ElementStrings:
        """Filter element strings by AI or data title.

        Args:
            ai: AI instance or string to match against the start of the
                element string's AI.
            data_title: String to find anywhere in the element string's AI
                data title.

        Returns:
            All matching element strings in the list.
        """
        if isinstance(ai, GS1ApplicationIdentifier):
            ai = ai.ai

        result = GS1ElementStrings()

        for element_string in self:
            ai_match = ai is not None and element_string.ai.ai.startswith(ai)
            data_title_match = (
                data_title is not None and data_title in element_string.ai.data_title
            )
            if ai_match or data_title_match:
                result.append(element_string)

        return result

    def get(
        self,
        *,
        ai: str | GS1ApplicationIdentifier | None = None,
        data_title: str | None = None,
    ) -> GS1ElementString | None:
        """Get element string by AI or data title.

        Args:
            ai: AI instance or string to match against the start of the
                element string's AI.
            data_title: String to find anywhere in the element string's AI
                data title.

        Returns:
            The first matching element string in the list.
        """
        matches = self.filter(ai=ai, data_title=data_title)
        return matches[0] if matches else None
